


/*
 * Companion program for pajpeg.spin2 
 *
 * Reference implementation of JPEG encoder
 * Generates unrolled loops for DCT and more
 * Converts tables into PASM compatible formats
 * Exports stats for statistical analysis
 *
 *
 * v 01 - initial working test
 * v 02 - screen door effect, dct broken
 * v 03 - took dct from toojpeg
 * v 04 - working half butterfly dct
 * v 05 - translating variable names for butterfly dct
 * v 06 - translating variable names for butterfly dct
 * v 07 - full butterfly dct, a few temp variables remain
 * v 08 - full butterfly dct
 * v 09 - bit reverse moved to end of dct
 * v 10 - renamed bitreverse to reorder, reorder in quant, not DCT
 * v 11 - dct uses flat array indexing instead of pass by reference, some code generation
 * v 12 - all dct operations abstracted, more code generation
 * v 13 - eliminate post-quant revserse scan
 * v 14 - pushed reorder into quant matrix
 * v 15 - messed*
 * v 16 - got ZigZagInv to work, quantizing in output order
 * v 17 - quant table in output order
 * v 18 - clean up bad code for table ordering
 * v 19 - end0pos analysis output, cleanup
 * v 20 - integer quantized dct reference
 * v 21 - integer quantizer
 * v 22 - good integer dct
 * v 24 - bugfix dct
 * v 25 - rebranch from v22
 * v 26 - good integer dct, bugfix
 * v 27 - unscaled dct test
 * v 28 - quantizer function
 * v 29 - quantizer uses arithmetic right shift
 * v 30 - uyvy input, 4:2:2 chroma subsampling
 * v 31 - fixed rgb input mode
 * v 32 - export headers and huffman tables 
 * v 33 - export scaled quant tables
 * v 34 - -128 shift moved to DCT
 * v 35 - code generator for Y input code, alt version
 * v 36 - code generator for Y and UV input regular version 
 * v 37 - added min/max value checker to DCT output
 * v 38 - removed rounding in DCT for bit-exact output with pajpeg11
 * v 39 - generate DCT multiplication constant table
 * v 40 - separate Y1 Y2 U V input blocks
 * v 41 - one subtract per DCT 8 for offset 
 *
 */



#define INPUT_UYVY


/* public domain Simple, Minimalistic JPEG writer - http://jonolick.com
 *
 * Quick Notes:
 * 	Based on a javascript jpeg writer
 * 	JPEG baseline (no JPEG progressive)
 * 	Supports 1, 3 or 4 component input. (luminance, RGB or RGBX)
 *
 * Latest revisions:
 *	1.52 (2012-22-11) Added support for specifying Luminance, RGB, or RGBA via comp(onents) argument (1, 3 and 4 respectively). 
 *	1.51 (2012-19-11) Fixed some warnings
 *	1.50 (2012-18-11) MT safe. Simplified. Optimized. Reduced memory requirements. Zero allocations. No namespace polution. Approx 340 lines code.
 *	1.10 (2012-16-11) compile fixes, added docs,
 *		changed from .h to .cpp (simpler to bootstrap), etc
 * 	1.00 (2012-02-02) initial release
 *
 * Basic usage:
 *	char *foo = new char[128*128*4]; // 4 component. RGBX format, where X is unused 
 *	jo_write_jpg("foo.jpg", foo, 128, 128, 4, 90); // comp can be 1, 3, or 4. Lum, RGB, or RGBX respectively.
 * 	
 * */

#ifndef JO_INCLUDE_JPEG_H
#define JO_INCLUDE_JPEG_H

// To get a header file for this, either cut and paste the header,
// or create jo_jpeg.h, #define JO_JPEG_HEADER_FILE_ONLY, and
// then include jo_jpeg.c from it.

typedef void jo_write_func(void *context, const void *data, int size);

// Returns false on failure
extern bool jo_write_jpg(const char *filename, const void *data, int width, int height, int comp, int quality);

// Returns false on failure
extern bool jo_write_jpg_to_func(jo_write_func *func, void *context, const void *data, int width, int height, int comp, int quality);

#endif // JO_INCLUDE_JPEG_H

#ifndef JO_JPEG_HEADER_FILE_ONLY

#if defined(_MSC_VER) && _MSC_VER >= 0x1400
#define _CRT_SECURE_NO_WARNINGS // suppress warnings about fopen()
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

static const unsigned char ZigZagInv[8*8] =
    {  0, 1, 8,16, 9, 2, 3,10,   // ZigZag[] =  0, 1, 5, 6,14,15,27,28,
      17,24,32,25,18,11, 4, 5,   //             2, 4, 7,13,16,26,29,42,
      12,19,26,33,40,48,41,34,   //             3, 8,12,17,25,30,41,43,
      27,20,13, 6, 7,14,21,28,   //             9,11,18,24,31,40,44,53,
      35,42,49,56,57,50,43,36,   //            10,19,23,32,39,45,52,54,
      29,22,15,23,30,37,44,51,   //            20,22,33,38,46,51,55,60,
      58,59,52,45,38,31,39,46,   //            21,34,37,47,50,56,59,61,
      53,60,61,54,47,55,62,63 }; //            35,36,48,49,57,58,62,63
static const unsigned char s_jo_ZigZag[] = { 0,1,5,6,14,15,27,28,2,4,7,13,16,26,29,42,3,8,12,17,25,30,41,43,9,11,18,24,31,40,44,53,10,19,23,32,39,45,52,54,20,22,33,38,46,51,55,60,21,34,37,47,50,56,59,61,35,36,48,49,57,58,62,63 };

static void jo_putc(jo_write_func *func, void *context, int c)
{
	func(context, (unsigned char *)&c, sizeof(unsigned char));
}

static void jo_writeBits(jo_write_func *func, void *context, int &bitBuf, int &bitCnt, const unsigned short *bs) {
	bitCnt += bs[1];
	bitBuf |= bs[0] << (24 - bitCnt);
	while(bitCnt >= 8) {
		unsigned char c = (bitBuf >> 16) & 255;
		jo_putc(func, context, c);
		if(c == 255) {
			jo_putc(func, context, 0);
		}
		bitBuf <<= 8;
		bitCnt -= 8;
	}
}
#define LANG_PASM2  (1)
int g_emit_language = LANG_PASM2;
#define EMIT_FILE stdout
char g_comment[128]={0,0};

static void emit_readline1( const char *name,  const int o )
{
	int base;
	int bindexa=2;
	int bindexb=0;
	if( g_emit_language==LANG_PASM2 )
	{
	fprintf(EMIT_FILE,"\n\n\n%s\n",name);
	for( base=0;base<64;base+=8)
	{
		char comment[128];
		sprintf(comment,"\t' %s unpack gray/planar line %2d",g_comment,base/8);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+7,o+base+1,3,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+6,o+base+1,2,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+5,o+base+1,1,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+4,o+base+1,0,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+3,o+base+0,3,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+2,o+base+0,2,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+1,o+base+0,1,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+0,o+base+0,0,comment);
	}
	//if(o!=0) // Y1 falls through to DCT
		fprintf(EMIT_FILE,"                jmp     #@hjpeg_dctquant\n");
	fprintf(EMIT_FILE,"\n\n\n\n");
	}
}
static void emit_readline_rgb( const char *name,  const int o )
{
	int base;
	int bindexa=2;
	int bindexb=0;
	int pos;
	int src_b=0;
	int dst_b;
	int pad=0;
	if (o==4)
		pad=1;
	if( g_emit_language==LANG_PASM2 )
	{
	fprintf(EMIT_FILE,"\n\n\n%s\n",name);
	for( base=0;base<64;base+=8)
	{
		src_b+=pad;
		char comment[128];
		for( pos=0; pos<8 ; pos++)
		{
		dst_b=pos+base;
		src_b=(base*4) + (o*pos) + pad;
		sprintf(comment,"\t' %s convert RGB%d line %d pixel %d",g_comment,o*8,base/8 ,pos);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",dst_b,src_b/4,src_b%4,comment);
		fprintf(EMIT_FILE,"                muls    d+%d,yuv_m+%d %s\n",dst_b,0,comment);

		src_b++;
		fprintf(EMIT_FILE,"                getbyte dcttmp,p+%d,#%d %s\n",src_b/4,src_b%4,comment);
		fprintf(EMIT_FILE,"                muls    dcttmp,yuv_m+%d %s\n",1,comment);
		fprintf(EMIT_FILE,"                add     d+%d,dcttmp   %s\n",dst_b,comment);

		src_b++;
		fprintf(EMIT_FILE,"                getbyte dcttmp,p+%d,#%d %s\n",src_b/4,src_b%4,comment);
		fprintf(EMIT_FILE,"                muls    dcttmp,yuv_m+%d %s\n",2,comment);
		fprintf(EMIT_FILE,"                add     d+%d,dcttmp   %s\n",dst_b,comment);



		//fprintf(EMIT_FILE,"                fges    d+%d,yuv_min   %s\n",dst_b,comment);
		//fprintf(EMIT_FILE,"                fles    d+%d,yuv_max   %s\n",dst_b,comment);
		fprintf(EMIT_FILE,"                sar     d+%d,yuv_sar   %s\n",dst_b,comment);
		}

	}
	if(o!=3) // Y1 falls through to DCT
		fprintf(EMIT_FILE,"                jmp     #@hjpeg_dctquant\n");
	fprintf(EMIT_FILE,"\n\n\n\n");
	}
}

static void emit_readline4( const char *name,  const int bindex )
{
	int base;
	//int bindex=0;
	if( g_emit_language==LANG_PASM2 )
	{
	fprintf(EMIT_FILE,"\n\n\n%s\n",name);
	for( base=0;base<64;base+=8)
	{
		char comment[128];
		sprintf(comment,"\t' %s unpack UYVY line %2d",g_comment,base/8);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+7,base+7,bindex,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+6,base+6,bindex,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+5,base+5,bindex,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+4,base+4,bindex,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+3,base+3,bindex,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+2,base+2,bindex,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+1,base+1,bindex,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+0,base+0,bindex,comment);
	}
		fprintf(EMIT_FILE,"                jmp     #@hjpeg_dctquant\n");
	fprintf(EMIT_FILE,"\n\n\n\n",name);
	}
}
static void emit_readline2( const char *name,  const int o )
{
	int base;
	int bindexa=2;
	int bindexb=0;
	if( g_emit_language==LANG_PASM2 )
	{
	fprintf(EMIT_FILE,"\n\n\n%s\n",name);
	for( base=0;base<64;base+=8)
	{
		char comment[128];
		sprintf(comment,"\t' %s unpack UYVY line %2d",g_comment,base/8);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+7,o+base+3,bindexa,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+6,o+base+3,bindexb,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+5,o+base+2,bindexa,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+4,o+base+2,bindexb,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+3,o+base+1,bindexa,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+2,o+base+1,bindexb,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+1,o+base+0,bindexa,comment);
		fprintf(EMIT_FILE,"                getbyte d+%d,p+%d,#%d   %s\n",base+0,o+base+0,bindexb,comment);
	}
	//if(o!=0) // Y1 falls through to DCT
		fprintf(EMIT_FILE,"                jmp     #@hjpeg_dctquant\n");
	fprintf(EMIT_FILE,"\n\n\n\n");
	}
}


static void shiftbutterflyar( int *data , int sum, int diff , int offset)
{
	int temp = data[sum];
	data[sum] += data[diff];
	data[sum] -=offset;
	data[diff] = temp - data[diff];
	if( g_emit_language==LANG_PASM2 )
	{
		char comment[128];
		sprintf(comment,"\t' %s sbfly(%2d,%2d)",g_comment,sum,diff);
		fprintf(EMIT_FILE,"                mov    dcttmp,d+%d %s\n",sum,comment);
		fprintf(EMIT_FILE,"                add    d+%d,d+%d   %s\n",sum,diff,comment);
		fprintf(EMIT_FILE,"                sub    d+%d,dctsub %s\n",sum,comment);
		fprintf(EMIT_FILE,"                subr   d+%d,dcttmp %s\n",diff,comment);
	}
}

static void butterflyar( int *data , int sum, int diff )
{
	int temp = data[sum];
	data[sum] += data[diff];
	data[diff] = temp - data[diff];
	if( g_emit_language==LANG_PASM2 )
	{
		char comment[128];
		sprintf(comment,"\t' %s  bfly(%2d,%2d)",g_comment,sum,diff);
		fprintf(EMIT_FILE,"                mov    dcttmp,d+%d %s\n",sum,comment);
		fprintf(EMIT_FILE,"                add    d+%d,d+%d   %s\n",sum,diff,comment);
		fprintf(EMIT_FILE,"                subr   d+%d,dcttmp %s\n",diff,comment);
	}
}

static void rbutterflyar( int *data , int sum, int diff )
{
	int temp = data[sum];
	data[sum] += data[diff];
	data[diff] = data[diff] - temp;
	if( g_emit_language==LANG_PASM2 )
	{
		char comment[128];
		sprintf(comment,"\t' %s rbfly(%2d,%2d)",g_comment,sum,diff);
		fprintf(EMIT_FILE,"                mov    dcttmp,d+%d %s\n",sum,comment);
		fprintf(EMIT_FILE,"                add    d+%d,d+%d   %s\n",sum,diff,comment);
		fprintf(EMIT_FILE,"                sub    d+%d,dcttmp %s\n",diff,comment);
	}
}

static void addar( int *data , int sum, int diff )
{
	data[sum] += data[diff];
	if( g_emit_language==LANG_PASM2 )
	{
		char comment[128];
		sprintf(comment,"\t' %s   add(%2d,%2d)",g_comment,sum,diff);
		fprintf(EMIT_FILE,"                add    d+%d,d+%d   %s\n",sum,diff,comment);
	}
}
static void subar( int *data , int sum, int diff )
{
	data[sum] -= data[diff];
	if( g_emit_language==LANG_PASM2 )
	{
		char comment[128];
		sprintf(comment,"\t' %s   sub(%2d,%2d)",g_comment,sum,diff);
		fprintf(EMIT_FILE,"                sub    d+%d,d+%d   %s\n",sum,diff,comment);
	}
}
static void movar( int *data , int sum, int diff )
{
	data[sum] = data[diff];
	if( g_emit_language==LANG_PASM2 )
	{
		char comment[128];
		sprintf(comment,"\t' %s   mov(%2d,%2d)",g_comment,sum,diff);
		fprintf(EMIT_FILE,"                mov    d+%d,d+%d   %s\n",sum,diff,comment);
	}
}

int dctm[16];
static void scalear( int *data , int val, float scale , int index)
{
	int imul=1<<14;
	int iscale = ( (scale* (float)imul) + 0.5 );
	dctm[index]=iscale;
	if ( data[val]>32767 ) printf("scale>16bits\n");
	if ( data[val]<-32767 ) printf("scale>16bits\n");
	//if (data[val]<0 )
	//	data[val] = (data[val]*iscale - imul/2)/imul;
	//else
	//	data[val] = (data[val]*iscale + imul/2)/imul;

	//data[val] = (data[val]*iscale + imul/2)>>14;
	data[val] = (data[val]*iscale)>>14;
	if( g_emit_language==LANG_PASM2 )
	{
		char comment[128];
		sprintf(comment,"\t' %s  scas(%2d,#%f)",g_comment,val,scale);
		//fprintf(EMIT_FILE,"                muls   d+%d,##%d   %s\n",val,iscale,comment);
		fprintf(EMIT_FILE,"                muls   d+%d,dctm+%d   %s\n",val,index,comment);
		fprintf(EMIT_FILE,"                sar    d+%d,#%d   %s\n",val,14,comment);
	}
}

static void quantar( int *qt, int *data , int qpos, int dpos )
{
	// will overwrite the data in qt[ qpos ]
	
	int qscale=1<<14;
	if ( data[dpos]>32767 ) printf("quant scale>16bits\n");
	if ( data[dpos]<-32767 ) printf("quant scale>16bits\n");
	qt[qpos] *= data[dpos];
	//if( data[dpos]<0 ) qt[qpos]-=qscale/2; else qt[qpos]+=qscale/2;
	//qt[qpos]/=qscale;	// this method has about 30 fewer differences with the floating point version
	
	qt[qpos]=((qscale/2)+qt[qpos])>>14; // arithmetic shift since we are using signed int

//	qt[qpos]+=((qscale>>1)); //
//	qt[qpos]>>=14; // arithmetic shift 
//	qt[qpos]>>=14; // arithmetic shift 

	if( g_emit_language==LANG_PASM2 )
	{
		char comment[128];
		sprintf(comment,"\t' %s  quant(%2d,%d)",g_comment,qpos,dpos);
		fprintf(EMIT_FILE,"                muls   qt+%d,d+%d      %s\n",qpos,dpos,comment);
		fprintf(EMIT_FILE,"                add    qt+%d,quantround%s\n",qpos,comment);
		fprintf(EMIT_FILE,"                sar    qt+%d,#%d   wz  %s\n",qpos,14,comment);
		if(qpos) // DC is handled separately
		fprintf(EMIT_FILE,"       if_nz    mov    end0pos,#%d     %s\n",qpos,comment);
		else
		sprintf(comment,"\t' %s clear end0pos",g_comment);
		fprintf(EMIT_FILE,"                mov    end0pos,#%d     %s\n",qpos,comment);

	}
}

static void butterfly( float &sum, float &diff )
{
	float temp = sum;
	sum += diff;
	//diff = sum - diff*2;
	diff = temp - diff;
}

static void rbutterfly( float &sum, float &diff )
{
	float temp = sum;
	sum += diff;
	//diff = sum - diff*2;
	diff = diff - temp;
}


const auto SqrtHalfSqrt = 1.306562965f; // sqrt((2 + sqrt(2)) / 2)  = cos(pi * 1 / 8) * sqrt(2)
const auto HalfSqrtSqrt = 0.382683432f; // sqrt( 2 - sqrt(2)) / 2   = cos(pi * 3 / 8)
const auto InvSqrt      = 0.707106781f; // 1 / sqrt(2)              = cos(pi * 2 / 8)
const auto InvSqrtSqrt  = 0.541196100f; // 1 / sqrt(2 - sqrt(2))    = cos(pi * 3 / 8) * sqrt(2)




static void bfly_DCTar(int *data , int start, int stride ) {
      

  int offset=128*8;
  if (stride==8)
  {
    sprintf(g_comment,"DCT col %d ",start);
  }
  else
  {
    sprintf(g_comment,"DCT row %d ",start/8);
  }

  butterflyar( data, start + stride*0 , start + stride*7);
  butterflyar( data, start + stride*1 , start + stride*6);
  butterflyar( data, start + stride*2 , start + stride*5);
  butterflyar( data, start + stride*3 , start + stride*4);

  butterflyar( data, start + stride*0 , start + stride*3);
  butterflyar( data, start + stride*1 , start + stride*2);

  if (stride==8){ // column
  butterflyar( data, start + stride*0 , start + stride*1);  
  }else{ // row (rows go first)
  shiftbutterflyar( data, start + stride*0 , start + stride*1, offset);  
  }

  addar( data, start + stride*2 , start + stride*3);  
  //data[ start + stride*2 ] = ( data[ start + stride*2 ] ) * InvSqrt; 
  scalear( data, start + stride*2 , InvSqrt, 2);  

  rbutterflyar( data, start + stride*2 , start + stride*3 );

  addar( data, start + stride*4 , start + stride*5);  
  addar( data, start + stride*5 , start + stride*6);  
  addar( data, start + stride*6 , start + stride*7);  

  //float z5 = (data[ start+ stride*4] - data[start + stride*6]) * HalfSqrtSqrt; // a5
  movar( data, 64 , start + stride*4); // [64] is z5
  subar( data, 64 , start + stride*6); // [64] is z5
  scalear( data, 64 , HalfSqrtSqrt , 0 ); // [64] is z5
  //float z5 = data[64];

  //data[start + stride*4] = data[start + stride*4] * InvSqrtSqrt  + z5; //a2
  scalear( data, start + stride*4 , InvSqrtSqrt, 1);  
  addar( data, start + stride*4 , 64 );  
  //data[start + stride*5] = data[start + stride*5] * InvSqrt; //a3
  scalear( data, start + stride*5 , InvSqrt, 2);  
  //data[start + stride*6] = data[start + stride*6] * SqrtHalfSqrt + z5; //a4
  scalear( data, start + stride*6 , SqrtHalfSqrt, 3);  
  addar( data, start + stride*6 , 64 );  

  rbutterflyar( data, start + stride*5 , start + stride*7 );
  rbutterflyar( data, start + stride*4 , start + stride*7 );
  butterflyar( data, start + stride*5 , start + stride*6 );

} 

static void bfly_DCT(float &d0, float &d1, float &d2, float &d3, float &d4, float &d5, float &d6, float &d7) {

  //float add07 = d0 + d7; float sub07 = d0 - d7; // tmp0, tmp7
  //float add16 = d1 + d6; float sub16 = d1 - d6; // tmp1, tmp6
  //float add25 = d2 + d5; float sub25 = d2 - d5; // tmp2, tmp5
  //float add34 = d3 + d4; float sub34 = d3 - d4; // tmp3, tmp4
  butterfly( d0 , d7);
  butterfly( d1 , d6);
  butterfly( d2 , d5);
  butterfly( d3 , d4);
  

  //float add0347 = add07 + add34; float sub07_34 = add07 - add34; // tmp10, tmp13 ("even part" / "phase 2")
  //float add1256 = add16 + add25; float sub16_25 = add16 - add25; // tmp11, tmp12
  butterfly( d0 , d3);
  butterfly( d1 , d2);


  //d0 = add0347 + add1256; d4 = add0347 - add1256; // "phase 3"
  butterfly( d0 , d1);  // d1 is really d4 (bit reversal)

  d2 += d3;
  //        d1-d2 + d3      a1
  d2 = ( d2 ) * InvSqrt; // all temporary z-variables kept their original names
  //float z1 = ( d2 + d3 ) * InvSqrt; // all temporary z-variables kept their original names
  //float z1 = (sub16_25 + sub07_34) * InvSqrt; // all temporary z-variables kept their original names
  //d2 = sub07_34 + z1; d6 = sub07_34 - z1; // "phase 5"
  //d2 = d3 + z1; d6 = d3 - z1; // "phase 5"
  rbutterfly( d2 , d3 );
  //float out2= d3 + z1;
  //float out6= d3 - z1;  


  d4 += d5;
  d5 += d6;
  d6 += d7;
  //                d2  + d3
  //float sub23_45 = d4;//d5+d4 ;//sub25 + sub34; // tmp10 ("odd part" / "phase 2")
  //float sub12_56 = d5;//d6+d5 ;//sub16 + sub25; // tmp11
  //float sub01_67 = d6;//d6+d7 ;//sub16 + sub07; // tmp12

  float z5 = (d4 - d6) * HalfSqrtSqrt; // a5
  //float z5 = (sub23_45 - sub01_67) * HalfSqrtSqrt; // a5
  d4 = d4 * InvSqrtSqrt  + z5; //a2
  //float z2 = d4 * InvSqrtSqrt  + z5; //a2
  //float z2 = sub23_45 * InvSqrtSqrt  + z5; //a2
  d5 = d5 * InvSqrt; //a3
  //float z3 = d5 * InvSqrt; //a3
  //float z3 = sub12_56 * InvSqrt; //a3
  d6 = d6 * SqrtHalfSqrt + z5; //a4
  //float z4 = d6 * SqrtHalfSqrt + z5; //a4
  //float z4 = sub01_67 * SqrtHalfSqrt + z5; //a4
					   //
  //float z6 = sub07 + z3; // z11 ("phase 5")
  //float z7 = sub07 - z3; // z13
  //float z6 = d7 + z3; // z11 ("phase 5")
  //float z7 = d7 - z3; // z13
  rbutterfly( d5 , d7 );
  //d1 = z6 + z4; d7 = z6 - z4; // "phase 6"
  //d5 = z7 + z2; d3 = z7 - z2;
  rbutterfly( d4 , d7 );
  butterfly( d5 , d6 );

  /*
  // reorder output
  float out0 = d0;  // 000 000 first 4 terms bitreversed
  float out4 = d1;  // 100 001
  float out2 = d2;  // 010 010
  float out6 = d3;  // 110 011
  float out5 = d4;  // 101 100 second 4 terms 
  float out1 = d5;  // 001 101
  float out7 = d6;  // 111 110
  float out3 = d7;  // 011 111

  d0 = out0;
  d1 = out1;
  d2 = out2;
  d3 = out3;
  d4 = out4;
  d5 = out5;
  d6 = out6;
  d7 = out7;
*/
} 
int invorderblock( int index )
{
	//int reorder[8] = { 0, 4, 2, 6, 5, 1, 7, 3};
	int reorder[8] = { 0, 5, 2, 7,  1, 4, 3, 6};

	int out=0;
	out = reorder[ index&7 ];
	out |= reorder[ (index>>3)&7 ]<<3;
	return out;
}
void emit_int_array( const char *name,  const int *array, int len )
{
	int i;
	printf("%s   long  ",name);
	for( i=0; i<len ; i++)
	{
		printf("%d%c",array[i],i==(len-1)?' ':',');
	}
	printf("   ' n=%d\n",len);
}
static void jo_DCT(float &d0, float &d1, float &d2, float &d3, float &d4, float &d5, float &d6, float &d7) {
	float tmp0 = d0 + d7;
	float tmp7 = d0 - d7;
	float tmp1 = d1 + d6;
	float tmp6 = d1 - d6;
	float tmp2 = d2 + d5;
	float tmp5 = d2 - d5;
	float tmp3 = d3 + d4;
	float tmp4 = d3 - d4;

	// Even part
	float tmp10 = tmp0 + tmp3;	// phase 2
	float tmp13 = tmp0 - tmp3;
	float tmp11 = tmp1 + tmp2;
	float tmp12 = tmp1 - tmp2;

	d0 = tmp10 + tmp11; 		// phase 3
	d4 = tmp10 - tmp11;

	float z1 = (tmp12 + tmp13) * 0.707106781f; // c4
	d2 = tmp13 + z1; 		// phase 5
	d6 = tmp13 - z1;

	// Odd part
	tmp10 = tmp4 + tmp5; 		// phase 2
	tmp11 = tmp5 + tmp6;
	tmp12 = tmp6 + tmp7;

	// The rotator is modified from fig 4-8 to avoid extra negations.
	float z5 = (tmp10 - tmp12) * 0.382683433f; // c6
	float z2 = tmp10 * 0.541196100f + z5; // c2-c6
	float z4 = tmp12 * 1.306562965f + z5; // c2+c6
	float z3 = tmp11 * 0.707106781f; // c4

	float z11 = tmp7 + z3;		// phase 5
	float z13 = tmp7 - z3;

	d5 = z13 + z2;			// phase 6
	d3 = z13 - z2;
	d1 = z11 + z4;
	d7 = z11 - z4;
} 

static void jo_calcBits(int val, unsigned short bits[2]) {
	int tmp1 = val < 0 ? -val : val;
	val = val < 0 ? val-1 : val;
	bits[1] = 1;
	while(tmp1 >>= 1) {
		++bits[1];
	}
	bits[0] = val & ((1<<bits[1])-1);
}
int reorderblock( int index )
{
	int reorder[8] = { 0, 4, 2, 6, 5, 1, 7, 3};

	int out=0;
	out = reorder[ index&7 ];
	out |= reorder[ (index>>3)&7 ]<<3;
	return out;
}

static int jo_processDU(jo_write_func *func, void *context, int &bitBuf, int &bitCnt, int *CDU, float *fdtbl, int DC, const unsigned short HTDC[256][2], const unsigned short HTAC[256][2]) {
	const unsigned short EOB[2] = { HTAC[0x00][0], HTAC[0x00][1] };
	const unsigned short M16zeroes[2] = { HTAC[0xF0][0], HTAC[0xF0][1] };

	emit_readline1("hjpeg_block_Y8",0);
	emit_readline4("hjpeg_block_U",1);
	emit_readline4("hjpeg_block_V",3);
	emit_readline2("hjpeg_block_Y2",4);
	emit_readline2("hjpeg_block_Y1",0);
	emit_readline_rgb("hjpeg_block_RGB32",4);
	emit_readline_rgb("hjpeg_block_RGB24",3);


	if( g_emit_language==LANG_PASM2 )
		fprintf(EMIT_FILE,"\n\n\n%s\n","hjpeg_dctquant");

	// DCT rows
	for(int dataOff=0; dataOff<64; dataOff+=8) {
		bfly_DCTar(CDU , dataOff, 1 ); 
	}
	// DCT columns
	for(int dataOff=0; dataOff<8; ++dataOff) {
		bfly_DCTar(CDU , dataOff, 8 ); 
	}
  
	sprintf(g_comment,"Quantization ");
	//int test=-20;
	//test+=8;
	//printf("test %d\n",test>>=4);

	// Quantize/descale/zigzag the coefficients
	int DU[64];
	int minv=1000000000, maxv=-1000000000;
	for(int i=0; i<64; ++i) {
		if( CDU[i]>maxv )  maxv=CDU[i];
		if( CDU[i]<minv )  minv=CDU[i];
		int qscale=1<<14;
		// load the quantization value
		DU[i] = (int) ( fdtbl[i]*(float)qscale + 0.5 ); // rounding does seem to improve performance
		quantar( DU, CDU, i, invorderblock( ZigZagInv[i])   );

		//int idctv = (int) (CDU[invorderblock(ZigZagInv[i])]);
		//if (idctv < 0 ) idctv-=2;
		//else idctv+=2;
		//idctv/=4;
		//int iqtv = (int) ( fdtbl[i]*(float)qscale + 0.5 ); // rounding does seem to improve performance
		//float v = (idctv * iqtv)/qscale; //int/int
		//float v = (idctv * fdtbl[i]); // quantized DCT only
		float v = CDU[invorderblock(ZigZagInv[i])]*fdtbl[i]/1.0; // float
		int fq = (int)(v < 0 ? ceilf(v - 0.5f) : floorf(v + 0.5f));
		//DU[(i)] = (int)(v < 0 ? ceilf(v - 0.5f) : floorf(v + 0.5f));
//		int iq = idctv < 0 ? (idctv*iqtv - qscale/2 )/qscale : (idctv*iqtv + qscale/2 )/qscale;
		//int iq = idctv < 0 ? (idctv*iqtv  )/8192 : (idctv*iqtv )/8192;
		//int iq = idctv < 0 ? (idctv*iqtv  )/8192 : (idctv*iqtv  )/8192; // significant size reduction!!
		int iq = DU[i];
		//DU[i]=iq;
		if( abs(iq-fq)>0 ) printf("%2d %f %3d %3d\n",i,1.0/fdtbl[i],iq,fq);
	}
		
	if( g_emit_language==LANG_PASM2 )
		fprintf(EMIT_FILE,"                jmp    #\\cjpeg_encode_mcu\n");

	g_emit_language=0; // emit 1 block only

	printf("min:%d max:%d\n",minv, maxv);

        emit_int_array( "dctm",  dctm, 6);

	// Encode DC
	int diff = DU[0] - DC; 
	if (diff == 0) {
		jo_writeBits(func, context, bitBuf, bitCnt, HTDC[0]);
	} else {
		unsigned short bits[2];
		jo_calcBits(diff, bits);
		jo_writeBits(func, context, bitBuf, bitCnt, HTDC[bits[1]]);
		jo_writeBits(func, context, bitBuf, bitCnt, bits);
		printf("diff:%x, HTIND:%x, bits:%x data%x\n",diff,bits[1],bits[0], HTDC[bits[1]][0]);
	}
	// Encode ACs
	int end0pos = 63;
	for(; (end0pos>0)&&(DU[end0pos]==0); --end0pos) {
	}
	// end0pos = first element in reverse order !=0
	printf("end0pos %d \n",end0pos);
	end0pos=63;
	if(end0pos == 0) {
		jo_writeBits(func, context, bitBuf, bitCnt, EOB);
		return DU[0];
	}
	int lastnz=0;
	for(int i = 1; i <= end0pos; ++i) {
		int startpos = i;
		for (; DU[i]==0 && i<=end0pos; ++i) {
		}
		//printf("i %d\n",i);
		if( i>=64 ) break;
		lastnz=i;
		int nrzeroes = i-startpos;
		if ( nrzeroes >= 16 ) {
			int lng = nrzeroes>>4;
			for (int nrmarker=1; nrmarker <= lng; ++nrmarker)
				jo_writeBits(func, context, bitBuf, bitCnt, M16zeroes);
			nrzeroes &= 15;
		}
		unsigned short bits[2];
		jo_calcBits(DU[i], bits);
		jo_writeBits(func, context, bitBuf, bitCnt, HTAC[(nrzeroes<<4)+bits[1]]);
		jo_writeBits(func, context, bitBuf, bitCnt, bits);
		printf("%d AC:%x, HTIND:%x, bits:%x data:%x\n",i,DU[i],(nrzeroes<<4)+bits[1],bits[0], HTAC[bits[1]+(nrzeroes<<4)][0]);
	}
	//printf("lastnz %d\n",lastnz);
	if(lastnz < 63) {
	//if(end0pos != 63) {
		jo_writeBits(func, context, bitBuf, bitCnt, EOB);
	}
	return DU[0];
}

static void jo_file_func(void *context, const void *data, int size)
{
	fwrite(data, size, 1, (FILE *)context);
}

bool jo_write_jpg(const char *filename, const void *data, int width, int height, int comp, int quality) {
	if(!filename) {
		return false;
	}

	FILE *fp = fopen(filename, "wb");
	if(!fp) {
		return false;
	}

	bool result = jo_write_jpg_to_func(jo_file_func, fp, data, width, height, comp, quality);
	fclose(fp);
	return result;
}
void emit_byte_array( const char *name, const unsigned char * array, int len )
{
	int i;
	printf("%s   byte  ",name);
	for( i=0; i<len ; i++)
	{
		printf("%d%c",array[i],i==(len-1)?' ':',');
	}
	printf("   ' n=%d\n",len);
}
void emit_long_array( const char *name,  const unsigned short array[256][2], int len )
{
	int i;
	unsigned int dw;
	printf("%s   long  ",name);
	//	printf("%d%c",array[0][0],array[0][1]);
	//	printf("%d%c",array[0],i==(len-1)?' ':',');
	for( i=0; i<len ; i++)
	{
		//printf("$%04x%04x%c",array[i][0],array[i][1],i==(len-1)?' ':','); orig
		printf("$%02x%06x%c",array[i][1],array[i][0],i==(len-1)?' ':',');
	}
	printf("   ' n=%d\n",len);
}


bool jo_write_jpg_to_func(jo_write_func *func, void *context, const void *data, int width, int height, int comp, int quality) {
	// Constants that don't pollute global namespace
	static const unsigned char std_dc_luminance_nrcodes[] = {0,0,1,5,1,1,1,1,1,1,0,0,0,0,0,0,0};
	static const unsigned char std_dc_luminance_values[] = {0,1,2,3,4,5,6,7,8,9,10,11};
	static const unsigned char std_ac_luminance_nrcodes[] = {0,0,2,1,3,3,2,4,3,5,5,4,4,0,0,1,0x7d};
	static const unsigned char std_ac_luminance_values[] = {
		0x01,0x02,0x03,0x00,0x04,0x11,0x05,0x12,0x21,0x31,0x41,0x06,0x13,0x51,0x61,0x07,0x22,0x71,0x14,0x32,0x81,0x91,0xa1,0x08,
		0x23,0x42,0xb1,0xc1,0x15,0x52,0xd1,0xf0,0x24,0x33,0x62,0x72,0x82,0x09,0x0a,0x16,0x17,0x18,0x19,0x1a,0x25,0x26,0x27,0x28,
		0x29,0x2a,0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x43,0x44,0x45,0x46,0x47,0x48,0x49,0x4a,0x53,0x54,0x55,0x56,0x57,0x58,0x59,
		0x5a,0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6a,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7a,0x83,0x84,0x85,0x86,0x87,0x88,0x89,
		0x8a,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9a,0xa2,0xa3,0xa4,0xa5,0xa6,0xa7,0xa8,0xa9,0xaa,0xb2,0xb3,0xb4,0xb5,0xb6,
		0xb7,0xb8,0xb9,0xba,0xc2,0xc3,0xc4,0xc5,0xc6,0xc7,0xc8,0xc9,0xca,0xd2,0xd3,0xd4,0xd5,0xd6,0xd7,0xd8,0xd9,0xda,0xe1,0xe2,
		0xe3,0xe4,0xe5,0xe6,0xe7,0xe8,0xe9,0xea,0xf1,0xf2,0xf3,0xf4,0xf5,0xf6,0xf7,0xf8,0xf9,0xfa
	};
	static const unsigned char std_dc_chrominance_nrcodes[] = {0,0,3,1,1,1,1,1,1,1,1,1,0,0,0,0,0};
	static const unsigned char std_dc_chrominance_values[] = {0,1,2,3,4,5,6,7,8,9,10,11};
	static const unsigned char std_ac_chrominance_nrcodes[] = {0,0,2,1,2,4,4,3,4,7,5,4,4,0,1,2,0x77};
	static const unsigned char std_ac_chrominance_values[] = {
		0x00,0x01,0x02,0x03,0x11,0x04,0x05,0x21,0x31,0x06,0x12,0x41,0x51,0x07,0x61,0x71,0x13,0x22,0x32,0x81,0x08,0x14,0x42,0x91,
		0xa1,0xb1,0xc1,0x09,0x23,0x33,0x52,0xf0,0x15,0x62,0x72,0xd1,0x0a,0x16,0x24,0x34,0xe1,0x25,0xf1,0x17,0x18,0x19,0x1a,0x26,
		0x27,0x28,0x29,0x2a,0x35,0x36,0x37,0x38,0x39,0x3a,0x43,0x44,0x45,0x46,0x47,0x48,0x49,0x4a,0x53,0x54,0x55,0x56,0x57,0x58,
		0x59,0x5a,0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6a,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7a,0x82,0x83,0x84,0x85,0x86,0x87,
		0x88,0x89,0x8a,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9a,0xa2,0xa3,0xa4,0xa5,0xa6,0xa7,0xa8,0xa9,0xaa,0xb2,0xb3,0xb4,
		0xb5,0xb6,0xb7,0xb8,0xb9,0xba,0xc2,0xc3,0xc4,0xc5,0xc6,0xc7,0xc8,0xc9,0xca,0xd2,0xd3,0xd4,0xd5,0xd6,0xd7,0xd8,0xd9,0xda,
		0xe2,0xe3,0xe4,0xe5,0xe6,0xe7,0xe8,0xe9,0xea,0xf2,0xf3,0xf4,0xf5,0xf6,0xf7,0xf8,0xf9,0xfa
	};
	// Huffman tables
	static const unsigned short YDC_HT[256][2] = { {0,2},{2,3},{3,3},{4,3},{5,3},{6,3},{14,4},{30,5},{62,6},{126,7},{254,8},{510,9}};
	static const unsigned short UVDC_HT[256][2] = { {0,2},{1,2},{2,2},{6,3},{14,4},{30,5},{62,6},{126,7},{254,8},{510,9},{1022,10},{2046,11}};
	static const unsigned short YAC_HT[256][2] = { 
		{10,4},{0,2},{1,2},{4,3},{11,4},{26,5},{120,7},{248,8},{1014,10},{65410,16},{65411,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{12,4},{27,5},{121,7},{502,9},{2038,11},{65412,16},{65413,16},{65414,16},{65415,16},{65416,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{28,5},{249,8},{1015,10},{4084,12},{65417,16},{65418,16},{65419,16},{65420,16},{65421,16},{65422,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{58,6},{503,9},{4085,12},{65423,16},{65424,16},{65425,16},{65426,16},{65427,16},{65428,16},{65429,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{59,6},{1016,10},{65430,16},{65431,16},{65432,16},{65433,16},{65434,16},{65435,16},{65436,16},{65437,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{122,7},{2039,11},{65438,16},{65439,16},{65440,16},{65441,16},{65442,16},{65443,16},{65444,16},{65445,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{123,7},{4086,12},{65446,16},{65447,16},{65448,16},{65449,16},{65450,16},{65451,16},{65452,16},{65453,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{250,8},{4087,12},{65454,16},{65455,16},{65456,16},{65457,16},{65458,16},{65459,16},{65460,16},{65461,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{504,9},{32704,15},{65462,16},{65463,16},{65464,16},{65465,16},{65466,16},{65467,16},{65468,16},{65469,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{505,9},{65470,16},{65471,16},{65472,16},{65473,16},{65474,16},{65475,16},{65476,16},{65477,16},{65478,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{506,9},{65479,16},{65480,16},{65481,16},{65482,16},{65483,16},{65484,16},{65485,16},{65486,16},{65487,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{1017,10},{65488,16},{65489,16},{65490,16},{65491,16},{65492,16},{65493,16},{65494,16},{65495,16},{65496,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{1018,10},{65497,16},{65498,16},{65499,16},{65500,16},{65501,16},{65502,16},{65503,16},{65504,16},{65505,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{2040,11},{65506,16},{65507,16},{65508,16},{65509,16},{65510,16},{65511,16},{65512,16},{65513,16},{65514,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{65515,16},{65516,16},{65517,16},{65518,16},{65519,16},{65520,16},{65521,16},{65522,16},{65523,16},{65524,16},{0,0},{0,0},{0,0},{0,0},{0,0},
		{2041,11},{65525,16},{65526,16},{65527,16},{65528,16},{65529,16},{65530,16},{65531,16},{65532,16},{65533,16},{65534,16},{0,0},{0,0},{0,0},{0,0},{0,0}
	};
	static const unsigned short UVAC_HT[256][2] = { 
		{0,2},{1,2},{4,3},{10,4},{24,5},{25,5},{56,6},{120,7},{500,9},{1014,10},{4084,12},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{11,4},{57,6},{246,8},{501,9},{2038,11},{4085,12},{65416,16},{65417,16},{65418,16},{65419,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{26,5},{247,8},{1015,10},{4086,12},{32706,15},{65420,16},{65421,16},{65422,16},{65423,16},{65424,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{27,5},{248,8},{1016,10},{4087,12},{65425,16},{65426,16},{65427,16},{65428,16},{65429,16},{65430,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{58,6},{502,9},{65431,16},{65432,16},{65433,16},{65434,16},{65435,16},{65436,16},{65437,16},{65438,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{59,6},{1017,10},{65439,16},{65440,16},{65441,16},{65442,16},{65443,16},{65444,16},{65445,16},{65446,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{121,7},{2039,11},{65447,16},{65448,16},{65449,16},{65450,16},{65451,16},{65452,16},{65453,16},{65454,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{122,7},{2040,11},{65455,16},{65456,16},{65457,16},{65458,16},{65459,16},{65460,16},{65461,16},{65462,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{249,8},{65463,16},{65464,16},{65465,16},{65466,16},{65467,16},{65468,16},{65469,16},{65470,16},{65471,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{503,9},{65472,16},{65473,16},{65474,16},{65475,16},{65476,16},{65477,16},{65478,16},{65479,16},{65480,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{504,9},{65481,16},{65482,16},{65483,16},{65484,16},{65485,16},{65486,16},{65487,16},{65488,16},{65489,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{505,9},{65490,16},{65491,16},{65492,16},{65493,16},{65494,16},{65495,16},{65496,16},{65497,16},{65498,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{506,9},{65499,16},{65500,16},{65501,16},{65502,16},{65503,16},{65504,16},{65505,16},{65506,16},{65507,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{2041,11},{65508,16},{65509,16},{65510,16},{65511,16},{65512,16},{65513,16},{65514,16},{65515,16},{65516,16},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
		{16352,14},{65517,16},{65518,16},{65519,16},{65520,16},{65521,16},{65522,16},{65523,16},{65524,16},{65525,16},{0,0},{0,0},{0,0},{0,0},{0,0},
		{1018,10},{32707,15},{65526,16},{65527,16},{65528,16},{65529,16},{65530,16},{65531,16},{65532,16},{65533,16},{65534,16},{0,0},{0,0},{0,0},{0,0},{0,0}
	};
	static const int YQT[] = {16,11,10,16,24,40,51,61,12,12,14,19,26,58,60,55,14,13,16,24,40,57,69,56,14,17,22,29,51,87,80,62,18,22,37,56,68,109,103,77,24,35,55,64,81,104,113,92,49,64,78,87,103,121,120,101,72,92,95,98,112,100,103,99};
	static const int UVQT[] = {17,18,24,47,99,99,99,99,18,21,26,66,99,99,99,99,24,26,56,99,99,99,99,99,47,66,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99};
	static const float aasf[] = { 1.0f * 2.828427125f, 1.387039845f * 2.828427125f, 1.306562965f * 2.828427125f, 1.175875602f * 2.828427125f, 1.0f * 2.828427125f, 0.785694958f * 2.828427125f, 0.541196100f * 2.828427125f, 0.275899379f * 2.828427125f };

	if(!data || !func || !context || !width || !height || comp > 4 || comp < 1 || comp == 2) {
		return false;
	}

	quality = quality ? quality : 90;
	quality = quality < 1 ? 1 : quality > 100 ? 100 : quality;
	quality = quality < 50 ? 5000 / quality : 200 - quality * 2;

	unsigned char YTable[64], UVTable[64];
	for(int i = 0; i < 64; ++i) {
		int yti = (YQT[ZigZagInv[i]]*quality+50)/100;
		YTable[i] = yti < 1 ? 1 : yti > 255 ? 255 : yti;
		int uvti  = (UVQT[ZigZagInv[i]]*quality+50)/100;
		UVTable[i] = uvti < 1 ? 1 : uvti > 255 ? 255 : uvti;
	}
	//Note: YTable and UVTable are stored in the image.
	//They are in output order. 
	float fdtbl_Y[64], fdtbl_UV[64];

	int  idtbl_Y[64], idtbl_UV[64];
	for(int k = 0; k <=63; ++k) 
	{
		int row=(ZigZagInv[k]>>3)&7;
		int col=(ZigZagInv[k]&7);
		fdtbl_Y[k]  = 1 / (YTable [k] * aasf[row] * aasf[col]);
		fdtbl_UV[k] = 1 / (UVTable[k] * aasf[row] * aasf[col]);
		int qscale=1<<14;
		idtbl_Y[k] = (int) ( fdtbl_Y[k]*(float)qscale + 0.5 ); // rounding does seem to improve performance
		idtbl_UV[k] = (int) ( fdtbl_UV[k]*(float)qscale + 0.5 ); 
	}
	// Quant tables in output order, but scaling factors need to be shuffled.
	emit_int_array("hjpeg_SQTY",idtbl_Y,64);
	emit_int_array("hjpeg_SQTUV",idtbl_UV,64);
	
	//g_emit_language=0;
	// Write Headers
	static const unsigned char head0[] = { 0xFF,0xD8,0xFF,0xE0,0,0x10,'J','F','I','F',0,1,1,0,0,1,0,1,0,0,0xFF,0xDB,0,0x84,0 };
	func(context, head0, sizeof(head0));
	func(context, YTable, sizeof(YTable));
	jo_putc(func, context, 1);
	func(context, UVTable, sizeof(UVTable));
#ifdef INPUT_UYVY
	const unsigned char head1[] = { 0xFF,0xC0,0,0x11,8,(unsigned char)(height>>8),(unsigned char)(height&0xFF),(unsigned char)(width>>8),(unsigned char)(width&0xFF),3,1,0x21,0,2,0x11,1,3,0x11,1,0xFF,0xC4,0x01,0xA2,0 };
	//const unsigned char head1[] = { 0xFF,0xC0,0,0x11,8,(unsigned char)(height>>8),(unsigned char)(height&0xFF),(unsigned char)(width>>8),(unsigned char)(width&0xFF),3,1,0x11,0,2,0x11,1,3,0x11,1,0xFF,0xC4,0x01,0xA2,0 };
#else
	const unsigned char head1[] = { 0xFF,0xC0,0,0x11,8,(unsigned char)(height>>8),(unsigned char)(height&0xFF),(unsigned char)(width>>8),(unsigned char)(width&0xFF),3,1,0x11,0,2,0x11,1,3,0x11,1,0xFF,0xC4,0x01,0xA2,0 };
#endif
	func(context, head1, sizeof(head1));
	func(context, std_dc_luminance_nrcodes+1, sizeof(std_dc_luminance_nrcodes)-1);
	func(context, std_dc_luminance_values, sizeof(std_dc_luminance_values));
	jo_putc(func, context, 0x10); // HTYACinfo
	func(context, std_ac_luminance_nrcodes+1, sizeof(std_ac_luminance_nrcodes)-1);
	func(context, std_ac_luminance_values, sizeof(std_ac_luminance_values));
	jo_putc(func, context, 1); // HTUDCinfo
	func(context, std_dc_chrominance_nrcodes+1, sizeof(std_dc_chrominance_nrcodes)-1);
	func(context, std_dc_chrominance_values, sizeof(std_dc_chrominance_values));
	jo_putc(func, context, 0x11); // HTUACinfo
	func(context, std_ac_chrominance_nrcodes+1, sizeof(std_ac_chrominance_nrcodes)-1);
	func(context, std_ac_chrominance_values, sizeof(std_ac_chrominance_values));
	static const unsigned char head2[] = { 0xFF,0xDA,0,0xC,3,1,0,2,0x11,3,0x11,0,0x3F,0 };
	func(context, head2, sizeof(head2));
	if (1)
	{
        const unsigned char ba[3]={0x10,1,0x11};
	emit_byte_array("YTable", YTable , 64 );
	emit_byte_array("UVTable", UVTable , 64 );
	emit_byte_array("std_dc_luminance_nrcodes", std_dc_luminance_nrcodes+1, sizeof(std_dc_luminance_nrcodes)-1);
	emit_byte_array("std_dc_luminance_values", std_dc_luminance_values, sizeof(std_dc_luminance_values));
	emit_byte_array("HTYACinfo", ba, 1);
	emit_byte_array("std_ac_luminance_nrcodes", std_ac_luminance_nrcodes+1, sizeof(std_ac_luminance_nrcodes)-1);
	emit_byte_array("std_ac_luminance_values", std_ac_luminance_values, sizeof(std_ac_luminance_values));
	emit_byte_array("HTUDCinfo", ba+1, 1);
	emit_byte_array("std_dc_chrominance_nrcodes", std_dc_chrominance_nrcodes+1, sizeof(std_dc_chrominance_nrcodes)-1);
	emit_byte_array("std_dc_chrominance_values", std_dc_chrominance_values, sizeof(std_dc_chrominance_values));
	emit_byte_array("HTUACinfo", ba+2, 1);
	emit_byte_array("std_ac_chrominance_nrcodes", std_ac_chrominance_nrcodes+1, sizeof(std_ac_chrominance_nrcodes)-1);
	emit_byte_array("std_ac_chrominance_values", std_ac_chrominance_values, sizeof(std_ac_chrominance_values));
	emit_long_array("YDC_HT", YDC_HT , 16 );
	emit_long_array("UVDC_HT", UVDC_HT , 16 );
	emit_long_array("YAC_HT", YAC_HT , 256 );
	emit_long_array("UVAC_HT", UVAC_HT , 256 );
	}

#ifdef INPUT_UYVY
	// Encode 8x8 macroblocks
	const unsigned char *imageData = (const unsigned char *)data;
	int DCY=0, DCU=0, DCV=0;
	int bitBuf=0, bitCnt=0;
	int ofsG = comp > 1 ? 1 : 0, ofsB = comp > 1 ? 2 : 0;
	for(int y = 0; y < height; y += 8) {
		for(int x = 0; x < width; x += 16) {
			int YDU[66], UDU[66], VDU[66], YDU2[66];
			for(int row = y, pos = 0; row < y+8; ++row) {
				for(int col = x; col < x+8; ++col, ++pos) {
					int p = row*width*2 + col*2;
					if(row >= height) {
						p -= width*2*(row+1 - height);
					}
					if(col >= width) {
						p -= 2*(col+1 - width);
					}
					YDU[pos]=imageData[p+0];
				}
			}
			for(int row = y, pos = 0; row < y+8; ++row) {
				for(int col = x+8; col < x+16; ++col, ++pos) {
					int p = row*width*2 + col*2;
					if(row >= height) {
						p -= width*2*(row+1 - height);
					}
					if(col >= width) {
						p -= 2*(col+1 - width);
					}
					YDU2[pos]=imageData[p+0];
				}
			}

			for(int row = y, pos = 0; row < y+8; ++row) {
				for(int col = x; col < x+16; col+=2, ++pos) {
					int p = row*width*2 + col*2;
					if(row >= height) {
						p -= width*2*(row+1 - height);
					}
					if(col >= width) {
						p -= 2*(col+1 - width);
					}
					UDU[pos]=imageData[p+1];
					VDU[pos]=imageData[p+3];
					//UDU[pos]=0;  VDU[pos]=0;
				}
			}

			DCY = jo_processDU(func, context, bitBuf, bitCnt, YDU, fdtbl_Y, DCY, YDC_HT, YAC_HT);
			DCY = jo_processDU(func, context, bitBuf, bitCnt, YDU2, fdtbl_Y, DCY, YDC_HT, YAC_HT);
			DCU = jo_processDU(func, context, bitBuf, bitCnt, UDU, fdtbl_UV, DCU, UVDC_HT, UVAC_HT);
			DCV = jo_processDU(func, context, bitBuf, bitCnt, VDU, fdtbl_UV, DCV, UVDC_HT, UVAC_HT);
		}// block loop x
	}// block loop y
#else
	// Encode 8x8 macroblocks
	const unsigned char *imageData = (const unsigned char *)data;
	int DCY=0, DCU=0, DCV=0;
	int bitBuf=0, bitCnt=0;
	int ofsG = comp > 1 ? 1 : 0, ofsB = comp > 1 ? 2 : 0;
	for(int y = 0; y < height; y += 8) {
		for(int x = 0; x < width; x += 8) {
			int YDU[66], UDU[66], VDU[66];
			for(int row = y, pos = 0; row < y+8; ++row) {
				for(int col = x; col < x+8; ++col, ++pos) {
					//int p = row*width*comp + col*comp;
					int p = row*width*comp + col*comp;
					if(row >= height) {
						p -= width*comp*(row+1 - height);
					}
					if(col >= width) {
						p -= comp*(col+1 - width);
					}

					float r = imageData[p+0], g = imageData[p+ofsG], b = imageData[p+ofsB];
					YDU[pos]=(+0.29900f*r+0.58700f*g+0.11400f*b-128)*1.0;
					UDU[pos]=(-0.16874f*r-0.33126f*g+0.50000f*b)*1.0;
					VDU[pos]=(+0.50000f*r-0.41869f*g-0.08131f*b)*1.0;
				}
			}

			DCY = jo_processDU(func, context, bitBuf, bitCnt, YDU, fdtbl_Y, DCY, YDC_HT, YAC_HT);
			DCU = jo_processDU(func, context, bitBuf, bitCnt, UDU, fdtbl_UV, DCU, UVDC_HT, UVAC_HT);
			DCV = jo_processDU(func, context, bitBuf, bitCnt, VDU, fdtbl_UV, DCV, UVDC_HT, UVAC_HT);
		}// block loop x
	}//block loop y
#endif	
	// Do the bit alignment of the EOI marker
	static const unsigned short fillBits[] = {0x7F, 7};
	jo_writeBits(func, context, bitBuf, bitCnt, fillBits);

	// EOI
	jo_putc(func, context, 0xFF);
	jo_putc(func, context, 0xD9);

	return true;
}

#ifdef INPUT_UYVY
#define IMG_H 240
#define IMG_W 320
#define IMG_BYTES 3
#define IMG_FILENAME "test.yuv"
#else
#define IMG_H 480
#define IMG_W 853
#define IMG_BYTES 3
#define IMG_FILENAME "testimg.rgb"
#endif
int main()
{

	char img[IMG_H*IMG_W*IMG_BYTES];
	FILE * f;
	f=fopen(IMG_FILENAME,"r");
	if( f )
	{
	fread( img, 1, IMG_H*IMG_W*IMG_BYTES , f);
	fclose(f);
	jo_write_jpg("joutput.jpg", img , IMG_W, IMG_H, IMG_BYTES, 90 );
	}

}
#endif

